// werapitest
//
// Exercises new Vista Windows Error Reporting API
//
// Claus Brod, http://www.clausbrod.de/Blog

#include <stdio.h>

#define _WIN32_WINNT 0x0500
#include <windows.h>
#include "werapi.h"

#include <atlbase.h>

#define FATAL_ERROR(s) do { OutputDebugStringA(s); DebugBreak(); } while (0)

typedef HRESULT (WINAPI *pfn_WERREPORTCREATE)(PCWSTR, WER_REPORT_TYPE, 
							  PWER_REPORT_INFORMATION, HREPORT *);
static pfn_WERREPORTCREATE pWerReportCreate = 0;

typedef HRESULT (WINAPI *pfn_WERREPORTSUBMIT)(HREPORT, WER_CONSENT, 
                                              DWORD, PWER_SUBMIT_RESULT);
static pfn_WERREPORTSUBMIT pWerReportSubmit = 0;

typedef HRESULT (WINAPI *pfn_WERREPORTCLOSEHANDLE)(HREPORT);
static pfn_WERREPORTCLOSEHANDLE pWerReportCloseHandle = 0;

typedef HRESULT (WINAPI *pfn_WERREPORTADDDUMP)(HREPORT, HANDLE, HANDLE, 
                                               WER_DUMP_TYPE, PWER_EXCEPTION_INFORMATION, 
                                               PWER_DUMP_CUSTOM_OPTIONS, DWORD);
static pfn_WERREPORTADDDUMP pWerReportAddDump = 0;


//--------- Launch debugger ----------------------------------------------------

TCHAR *installed_debugger()
{
  CRegKey key;
  if (ERROR_SUCCESS != key.Open(HKEY_LOCAL_MACHINE, 
    "Software\\Microsoft\\Windows NT\\CurrentVersion\\AeDebug\\", KEY_READ))
    return NULL;

  static TCHAR debuggerPath[1024];
  debuggerPath[0] = 0;
  DWORD dwCount = _countof(debuggerPath);
  key.QueryStringValue("Debugger", debuggerPath, &dwCount);
  return debuggerPath[0] ? debuggerPath : NULL;
}

static bool launch_debugger(void)
{
  TCHAR *debugger = installed_debugger();
  if (!debugger)
    return false;

  bool debuggerLaunched = false;

  // create event
  SECURITY_ATTRIBUTES attributes = { sizeof(attributes), 0, TRUE };
  HANDLE debugEvent = CreateEvent(&attributes, TRUE, FALSE, NULL);
  if (debugEvent) {
    // get process ID
    DWORD pid = GetCurrentProcessId();

    // build command string
    TCHAR commandBuffer[1024];
    _sntprintf_s(commandBuffer, _countof(commandBuffer), _TRUNCATE,
      debugger, pid, debugEvent);

    // start JIT debugger
    STARTUPINFO sinfo;
    memset(&sinfo, 0, sizeof(sinfo));
    sinfo.cb = sizeof(sinfo);

    PROCESS_INFORMATION pinfo;
    BOOL createProcRet = CreateProcess(NULL, commandBuffer, 
      NULL, NULL, 
      TRUE,  // inherit (event) handles
      0, NULL, NULL, &sinfo, &pinfo);

    if (createProcRet) {
      // wait until debugger fires event
      DWORD dwWaitResult = 0;
      dwWaitResult = WaitForSingleObject(debugEvent, 1000*100);
      if (dwWaitResult == WAIT_OBJECT_0) {
        debuggerLaunched = true;
      }
    } else {
      FATAL_ERROR("Could not run debugger process.\n");
    }
  }

  return debuggerLaunched;
}


//--------- Report crash using WER APIs ----------------------------------------

static bool report_crash(_EXCEPTION_POINTERS *inExceptionPointer)
{
  // Set up parameters for WerReportCreate()
  WER_REPORT_INFORMATION werReportInfo;
  memset(&werReportInfo, 0, sizeof(werReportInfo));
  werReportInfo.dwSize = sizeof(werReportInfo);
  wcscpy_s(werReportInfo.wzFriendlyEventName, 
    _countof(werReportInfo.wzFriendlyEventName), 
	L"werapitest (friendly event name)");
  wcscpy_s(werReportInfo.wzApplicationName, 
    _countof(werReportInfo.wzApplicationName), L"");
  wcscpy_s(werReportInfo.wzDescription, 
    _countof(werReportInfo.wzDescription), L"Critical runtime problem");

  PCWSTR eventType = L"werapitest (eventType)"; // APPCRASH
  HREPORT hReportHandle;
  if (FAILED(pWerReportCreate(eventType, WerReportCritical,
    &werReportInfo, &hReportHandle)) || !hReportHandle) {
      return false;
  }

  bool ret = false;

  WER_EXCEPTION_INFORMATION werExceptionInformation;
  werExceptionInformation.bClientPointers = FALSE;
  werExceptionInformation.pExceptionPointers = inExceptionPointer;
  bool dumpAdded = SUCCEEDED(pWerReportAddDump(hReportHandle, ::GetCurrentProcess(),
    ::GetCurrentThread(), WerDumpTypeMiniDump, &werExceptionInformation, NULL, 0));
  if (!dumpAdded) {
    FATAL_ERROR("Minidump generation failed.\n");
  }

  DWORD submitOptions = WER_SUBMIT_OUTOFPROCESS | WER_SUBMIT_NO_CLOSE_UI;
  WER_SUBMIT_RESULT submitResult;
  if (SUCCEEDED(pWerReportSubmit(hReportHandle, WerConsentNotAsked,
    submitOptions, &submitResult))) {
      switch(submitResult)
      {
      case WerDisabled:
        printf("Error reporting was disabled.\n");
        break;
      case WerDisabledQueue:
        printf("Queuing was disabled.\n");
        break;
      case WerReportAsync:
        printf("Report was asynchronous.\n");
        ret = true;
        break;
      case WerReportCancelled:
        printf("Report was cancelled.\n");
        break;
      case WerReportDebug:
        printf("Debug button clicked.\n");
        if (launch_debugger()) {
          printf("Debugger launched successfully.\n");
          DebugBreak();
          ret = true;
        }
        break;
      case WerReportFailed:
        printf("Report submission failed.\n");
        break;
      case WerReportQueued:
        printf("Report queued.\n");
        ret = true;
        break;
      case WerReportUploaded:
        printf("Report uploaded.\n");
        ret = true;
        break;
      default:
        printf("Unexpected return value %d from WerReportSubmit().\n",
          submitResult);
        break;
      }
  }
  pWerReportCloseHandle(hReportHandle);

  return ret;
}

static int filter_exception(EXCEPTION_POINTERS *exc_ptr)
{
  report_crash(exc_ptr);
  return EXCEPTION_EXECUTE_HANDLER;
}

static void wedding_crasher(void)
{
  __try {
    int *foo = (int *)0;
    *foo = 42;
  } __except(filter_exception(GetExceptionInformation())) {
    printf("Now in exception handler, process is still alive!\n");
  }
  Sleep(5000);
}

int main()
{
  HMODULE hWer = LoadLibrary("Wer.dll");
  if (hWer) {
    pWerReportCreate = 
      (pfn_WERREPORTCREATE)GetProcAddress(hWer, "WerReportCreate");
    pWerReportSubmit = 
      (pfn_WERREPORTSUBMIT)GetProcAddress(hWer, "WerReportSubmit");
    pWerReportCloseHandle = 
      (pfn_WERREPORTCLOSEHANDLE)GetProcAddress(hWer, "WerReportCloseHandle");
    pWerReportAddDump = 
      (pfn_WERREPORTADDDUMP)GetProcAddress(hWer, "WerReportAddDump");
  }

  if (!pWerReportCreate || !pWerReportSubmit || 
    !pWerReportCloseHandle || !pWerReportAddDump) {
      printf("Cannot initialize WER API.\n");
      return 1;
  }

  wedding_crasher();
  return 0;
}
